import time
from unittest import TestCase
from unittest import mock

from authlib.integrations.requests_client import AssertionSession


class AssertionSessionTest(TestCase):
    def setUp(self):
        self.token = {
            "token_type": "Bearer",
            "access_token": "a",
            "refresh_token": "b",
            "expires_in": "3600",
            "expires_at": int(time.time()) + 3600,
        }

    def test_refresh_token(self):
        def verifier(r, **kwargs):
            resp = mock.MagicMock()
            resp.status_code = 200
            if r.url == "https://i.b/token":
                self.assertIn("assertion=", r.body)
                resp.json = lambda: self.token
            return resp

        sess = AssertionSession(
            "https://i.b/token",
            issuer="foo",
            subject="foo",
            audience="foo",
            alg="HS256",
            key="secret",
        )
        sess.send = verifier
        sess.get("https://i.b")

        # trigger more case
        now = int(time.time())
        sess = AssertionSession(
            "https://i.b/token",
            issuer="foo",
            subject=None,
            audience="foo",
            issued_at=now,
            expires_at=now + 3600,
            header={"alg": "HS256"},
            key="secret",
            scope="email",
            claims={"test_mode": "true"},
        )
        sess.send = verifier
        sess.get("https://i.b")
        # trigger for branch test case
        sess.get("https://i.b")

    def test_without_alg(self):
        sess = AssertionSession(
            "https://i.b/token",
            grant_type=AssertionSession.JWT_BEARER_GRANT_TYPE,
            issuer="foo",
            subject="foo",
            audience="foo",
            key="secret",
        )
        self.assertRaises(ValueError, sess.get, "https://i.b")



import json
import os
import unittest

from cryptography.hazmat.primitives.keywrap import InvalidUnwrap

from authlib.common.encoding import json_b64encode
from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose import JsonWebEncryption
from authlib.jose import OctKey
from authlib.jose import OKPKey
from authlib.jose import errors
from authlib.jose.drafts import register_jwe_draft
from authlib.jose.errors import DecodeError
from authlib.jose.errors import InvalidAlgorithmForMultipleRecipientsMode
from authlib.jose.errors import InvalidHeaderParameterNameError
from authlib.jose.util import extract_header
from tests.util import read_file_path

register_jwe_draft(JsonWebEncryption)


class JWETest(unittest.TestCase):
    def test_not_enough_segments(self):
        s = "a.b.c"
        jwe = JsonWebEncryption()
        self.assertRaises(errors.DecodeError, jwe.deserialize_compact, s, None)

    def test_invalid_header(self):
        jwe = JsonWebEncryption()
        public_key = read_file_path("rsa_public.pem")
        self.assertRaises(
            errors.MissingAlgorithmError, jwe.serialize_compact, {}, "a", public_key
        )
        self.assertRaises(
            errors.UnsupportedAlgorithmError,
            jwe.serialize_compact,
            {"alg": "invalid"},
            "a",
            public_key,
        )
        self.assertRaises(
            errors.MissingEncryptionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP"},
            "a",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedEncryptionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP", "enc": "invalid"},
            "a",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedCompressionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP", "enc": "A256GCM", "zip": "invalid"},
            "a",
            public_key,
        )

    def test_not_supported_alg(self):
        public_key = read_file_path("rsa_public.pem")
        private_key = read_file_path("rsa_private.pem")

        jwe = JsonWebEncryption()
        s = jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A256GCM"}, "hello", public_key
        )

        jwe = JsonWebEncryption(algorithms=["RSA1_5", "A256GCM"])
        self.assertRaises(
            errors.UnsupportedAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP", "enc": "A256GCM"},
            "hello",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedCompressionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA1_5", "enc": "A256GCM", "zip": "DEF"},
            "hello",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedAlgorithmError,
            jwe.deserialize_compact,
            s,
            private_key,
        )

        jwe = JsonWebEncryption(algorithms=["RSA-OAEP", "A192GCM"])
        self.assertRaises(
            errors.UnsupportedEncryptionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP", "enc": "A256GCM"},
            "hello",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedCompressionAlgorithmError,
            jwe.serialize_compact,
            {"alg": "RSA-OAEP", "enc": "A192GCM", "zip": "DEF"},
            "hello",
            public_key,
        )
        self.assertRaises(
            errors.UnsupportedEncryptionAlgorithmError,
            jwe.deserialize_compact,
            s,
            private_key,
        )

    def test_inappropriate_sender_key_for_serialize_compact(self):
        jwe = JsonWebEncryption()
        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}
        self.assertRaises(
            ValueError, jwe.serialize_compact, protected, b"hello", bob_key
        )

        protected = {"alg": "ECDH-ES", "enc": "A256GCM"}
        self.assertRaises(
            ValueError,
            jwe.serialize_compact,
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    def test_inappropriate_sender_key_for_deserialize_compact(self):
        jwe = JsonWebEncryption()
        alice_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
            "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
            "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
        }
        bob_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}
        data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
        self.assertRaises(ValueError, jwe.deserialize_compact, data, bob_key)

        protected = {"alg": "ECDH-ES", "enc": "A256GCM"}
        data = jwe.serialize_compact(protected, b"hello", bob_key)
        self.assertRaises(
            ValueError, jwe.deserialize_compact, data, bob_key, sender_key=alice_key
        )

    def test_compact_rsa(self):
        jwe = JsonWebEncryption()
        s = jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A256GCM"},
            "hello",
            read_file_path("rsa_public.pem"),
        )
        data = jwe.deserialize_compact(s, read_file_path("rsa_private.pem"))
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "RSA-OAEP")

    def test_with_zip_header(self):
        jwe = JsonWebEncryption()
        s = jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A128CBC-HS256", "zip": "DEF"},
            "hello",
            read_file_path("rsa_public.pem"),
        )
        data = jwe.deserialize_compact(s, read_file_path("rsa_private.pem"))
        header, payload = data["header"], data["payload"]
        self.assertEqual(payload, b"hello")
        self.assertEqual(header["alg"], "RSA-OAEP")

    def test_aes_jwe(self):
        jwe = JsonWebEncryption()
        sizes = [128, 192, 256]
        _enc_choices = [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]
        for s in sizes:
            alg = f"A{s}KW"
            key = os.urandom(s // 8)
            for enc in _enc_choices:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(protected, b"hello", key)
                rv = jwe.deserialize_compact(data, key)
                self.assertEqual(rv["payload"], b"hello")

    def test_aes_jwe_invalid_key(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "A128KW", "enc": "A128GCM"}
        self.assertRaises(
            ValueError, jwe.serialize_compact, protected, b"hello", b"invalid-key"
        )

    def test_aes_gcm_jwe(self):
        jwe = JsonWebEncryption()
        sizes = [128, 192, 256]
        _enc_choices = [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]
        for s in sizes:
            alg = f"A{s}GCMKW"
            key = os.urandom(s // 8)
            for enc in _enc_choices:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(protected, b"hello", key)
                rv = jwe.deserialize_compact(data, key)
                self.assertEqual(rv["payload"], b"hello")

    def test_aes_gcm_jwe_invalid_key(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "A128GCMKW", "enc": "A128GCM"}
        self.assertRaises(
            ValueError, jwe.serialize_compact, protected, b"hello", b"invalid-key"
        )

    def test_serialize_compact_fails_if_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}

        self.assertRaises(
            InvalidHeaderParameterNameError,
            jwe.serialize_compact,
            protected,
            b"hello",
            key,
        )

    def test_serialize_compact_allows_unknown_fields_in_header_while_private_fields_not_restricted(
        self,
    ):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}

        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_serialize_json_fails_if_protected_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}
        header_obj = {"protected": protected}

        self.assertRaises(
            InvalidHeaderParameterNameError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            key,
        )

    def test_serialize_json_fails_if_unprotected_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        unprotected = {"foo": "bar"}
        header_obj = {"protected": protected, "unprotected": unprotected}

        self.assertRaises(
            InvalidHeaderParameterNameError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            key,
        )

    def test_serialize_json_fails_if_recipient_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        recipients = [{"header": {"foo": "bar"}}]
        header_obj = {"protected": protected, "recipients": recipients}

        self.assertRaises(
            InvalidHeaderParameterNameError,
            jwe.serialize_json,
            header_obj,
            b"hello",
            key,
        )

    def test_serialize_json_allows_unknown_fields_in_headers_while_private_fields_not_restricted(
        self,
    ):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo1": "bar1"}
        unprotected = {"foo2": "bar2"}
        recipients = [{"header": {"foo3": "bar3"}}]
        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
        }

        data = jwe.serialize_json(header_obj, b"hello", key)
        rv = jwe.deserialize_json(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_serialize_json_ignores_additional_members_in_recipients_elements(self):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}

        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_deserialize_json_fails_if_protected_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        decoded_protected = extract_header(to_bytes(data["protected"]), DecodeError)
        decoded_protected["foo"] = "bar"
        data["protected"] = to_unicode(json_b64encode(decoded_protected))

        self.assertRaises(
            InvalidHeaderParameterNameError, jwe.deserialize_json, data, key
        )

    def test_deserialize_json_fails_if_unprotected_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        data["unprotected"] = {"foo": "bar"}

        self.assertRaises(
            InvalidHeaderParameterNameError, jwe.deserialize_json, data, key
        )

    def test_deserialize_json_fails_if_recipient_header_contains_unknown_field_while_private_fields_restricted(
        self,
    ):
        jwe = JsonWebEncryption(private_headers=set())
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        data["recipients"][0]["header"] = {"foo": "bar"}

        self.assertRaises(
            InvalidHeaderParameterNameError, jwe.deserialize_json, data, key
        )

    def test_deserialize_json_allows_unknown_fields_in_headers_while_private_fields_not_restricted(
        self,
    ):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        data["unprotected"] = {"foo1": "bar1"}
        data["recipients"][0]["header"] = {"foo2": "bar2"}

        rv = jwe.deserialize_json(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_deserialize_json_ignores_additional_members_in_recipients_elements(self):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        data["recipients"][0]["foo"] = "bar"

        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_deserialize_json_ignores_additional_members_in_jwe_message(self):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
        header_obj = {"protected": protected}

        data = jwe.serialize_json(header_obj, b"hello", key)

        data["foo"] = "bar"

        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_key_agreement_computation(self):
        # https://tools.ietf.org/html/rfc7518#appendix-C
        alice_ephemeral_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
            "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
            "d": "0_NxaRPUMQoAJt50Gz8YiTr8gRTwyEaCumd-MToTmIo",
        }
        bob_static_key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        headers = {
            "alg": "ECDH-ES",
            "enc": "A128GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
            "epk": {
                "kty": "EC",
                "crv": "P-256",
                "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
                "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
            },
        }

        alg = JsonWebEncryption.ALG_REGISTRY["ECDH-ES"]
        enc = JsonWebEncryption.ENC_REGISTRY["A128GCM"]

        alice_ephemeral_key = alg.prepare_key(alice_ephemeral_key)
        bob_static_key = alg.prepare_key(bob_static_key)

        alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")
        bob_static_pubkey = bob_static_key.get_op_key("wrapKey")

        # Derived key computation at Alice

        # Step-by-step methods verification
        _shared_key_at_alice = alice_ephemeral_key.exchange_shared_key(
            bob_static_pubkey
        )
        self.assertEqual(
            _shared_key_at_alice,
            bytes(
                [
                    158,
                    86,
                    217,
                    29,
                    129,
                    113,
                    53,
                    211,
                    114,
                    131,
                    66,
                    131,
                    191,
                    132,
                    38,
                    156,
                    251,
                    49,
                    110,
                    163,
                    218,
                    128,
                    106,
                    72,
                    246,
                    218,
                    167,
                    121,
                    140,
                    254,
                    144,
                    196,
                ]
            ),
        )

        _fixed_info_at_alice = alg.compute_fixed_info(headers, enc.key_size)
        self.assertEqual(
            _fixed_info_at_alice,
            bytes(
                [
                    0,
                    0,
                    0,
                    7,
                    65,
                    49,
                    50,
                    56,
                    71,
                    67,
                    77,
                    0,
                    0,
                    0,
                    5,
                    65,
                    108,
                    105,
                    99,
                    101,
                    0,
                    0,
                    0,
                    3,
                    66,
                    111,
                    98,
                    0,
                    0,
                    0,
                    128,
                ]
            ),
        )

        _dk_at_alice = alg.compute_derived_key(
            _shared_key_at_alice, _fixed_info_at_alice, enc.key_size
        )
        self.assertEqual(
            _dk_at_alice,
            bytes(
                [86, 170, 141, 234, 248, 35, 109, 32, 92, 34, 40, 205, 113, 167, 16, 26]
            ),
        )
        self.assertEqual(urlsafe_b64encode(_dk_at_alice), b"VqqN6vgjbSBcIijNcacQGg")

        # All-in-one method verification
        dk_at_alice = alg.deliver(
            alice_ephemeral_key, bob_static_pubkey, headers, enc.key_size
        )
        self.assertEqual(
            dk_at_alice,
            bytes(
                [86, 170, 141, 234, 248, 35, 109, 32, 92, 34, 40, 205, 113, 167, 16, 26]
            ),
        )
        self.assertEqual(urlsafe_b64encode(dk_at_alice), b"VqqN6vgjbSBcIijNcacQGg")

        # Derived key computation at Bob

        # Step-by-step methods verification
        _shared_key_at_bob = bob_static_key.exchange_shared_key(alice_ephemeral_pubkey)
        self.assertEqual(_shared_key_at_bob, _shared_key_at_alice)

        _fixed_info_at_bob = alg.compute_fixed_info(headers, enc.key_size)
        self.assertEqual(_fixed_info_at_bob, _fixed_info_at_alice)

        _dk_at_bob = alg.compute_derived_key(
            _shared_key_at_bob, _fixed_info_at_bob, enc.key_size
        )
        self.assertEqual(_dk_at_bob, _dk_at_alice)

        # All-in-one method verification
        dk_at_bob = alg.deliver(
            bob_static_key, alice_ephemeral_pubkey, headers, enc.key_size
        )
        self.assertEqual(dk_at_bob, dk_at_alice)

    def test_ecdh_es_jwe_in_direct_key_agreement_mode(self):
        jwe = JsonWebEncryption()
        key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": "ECDH-ES", "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_jwe_json_serialization_single_recipient_in_direct_key_agreement_mode(
        self,
    ):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES", "enc": "A128GCM"}
        header_obj = {"protected": protected}
        data = jwe.serialize_json(header_obj, b"hello", key)
        rv = jwe.deserialize_json(data, key)
        self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_jwe_in_key_agreement_with_key_wrapping_mode(self):
        jwe = JsonWebEncryption()
        key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
            "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
        }

        for alg in [
            "ECDH-ES+A128KW",
            "ECDH-ES+A192KW",
            "ECDH-ES+A256KW",
        ]:
            for enc in [
                "A128CBC-HS256",
                "A192CBC-HS384",
                "A256CBC-HS512",
                "A128GCM",
                "A192GCM",
                "A256GCM",
            ]:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(protected, b"hello", key)
                rv = jwe.deserialize_compact(data, key)
                self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_jwe_with_okp_key_in_direct_key_agreement_mode(self):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": "ECDH-ES", "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_jwe_with_okp_key_in_key_agreement_with_key_wrapping_mode(self):
        jwe = JsonWebEncryption()
        key = OKPKey.generate_key("X25519", is_private=True)

        for alg in [
            "ECDH-ES+A128KW",
            "ECDH-ES+A192KW",
            "ECDH-ES+A256KW",
        ]:
            for enc in [
                "A128CBC-HS256",
                "A192CBC-HS384",
                "A256CBC-HS512",
                "A128GCM",
                "A192GCM",
                "A256GCM",
            ]:
                protected = {"alg": alg, "enc": enc}
                data = jwe.serialize_compact(protected, b"hello", key)
                rv = jwe.deserialize_compact(data, key)
                self.assertEqual(rv["payload"], b"hello")

    def test_ecdh_es_jwe_with_json_serialization_when_kid_is_not_specified(self):
        jwe = JsonWebEncryption()

        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-ES+A256KW",
            "enc": "A256GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {"header": {"kid": "bob-key-2"}},
            {"header": {"kid": "2021-05-06"}},
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, [bob_key, charlie_key])

        rv_at_bob = jwe.deserialize_json(data, bob_key)

        self.assertEqual(
            rv_at_bob["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_bob["header"]["protected"][k]
                for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_bob["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_bob["header"]["recipients"], recipients)
        self.assertEqual(rv_at_bob["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_bob["payload"], payload)

        rv_at_charlie = jwe.deserialize_json(data, charlie_key)

        self.assertEqual(
            rv_at_charlie["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_charlie["header"]["protected"][k]
                for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_charlie["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_charlie["header"]["recipients"], recipients)
        self.assertEqual(rv_at_charlie["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_charlie["payload"], payload)

    def test_ecdh_es_jwe_with_json_serialization_when_kid_is_specified(self):
        jwe = JsonWebEncryption()

        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "bob-key-2",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "kid": "2021-05-06",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-ES+A256KW",
            "enc": "A256GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [
            {"header": {"kid": "bob-key-2"}},
            {"header": {"kid": "2021-05-06"}},
        ]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, [bob_key, charlie_key])

        rv_at_bob = jwe.deserialize_json(data, bob_key)

        self.assertEqual(
            rv_at_bob["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_bob["header"]["protected"][k]
                for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_bob["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_bob["header"]["recipients"], recipients)
        self.assertEqual(rv_at_bob["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_bob["payload"], payload)

        rv_at_charlie = jwe.deserialize_json(data, charlie_key)

        self.assertEqual(
            rv_at_charlie["header"]["protected"].keys(), protected.keys() | {"epk"}
        )
        self.assertEqual(
            {
                k: rv_at_charlie["header"]["protected"][k]
                for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv_at_charlie["header"]["unprotected"], unprotected)
        self.assertEqual(rv_at_charlie["header"]["recipients"], recipients)
        self.assertEqual(rv_at_charlie["header"]["aad"], jwe_aad)
        self.assertEqual(rv_at_charlie["payload"], payload)

    def test_ecdh_es_jwe_with_json_serialization_for_single_recipient(self):
        jwe = JsonWebEncryption()

        key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )

        protected = {
            "alg": "ECDH-ES+A256KW",
            "enc": "A256GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [{"header": {"kid": "bob-key-2"}}]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, key)

        rv = jwe.deserialize_json(data, key)

        self.assertEqual(rv["header"]["protected"].keys(), protected.keys() | {"epk"})
        self.assertEqual(
            {
                k: rv["header"]["protected"][k]
                for k in rv["header"]["protected"].keys() - {"epk"}
            },
            protected,
        )
        self.assertEqual(rv["header"]["unprotected"], unprotected)
        self.assertEqual(rv["header"]["recipients"], recipients)
        self.assertEqual(rv["header"]["aad"], jwe_aad)
        self.assertEqual(rv["payload"], payload)

    def test_ecdh_es_encryption_fails_json_serialization_multiple_recipients_in_direct_key_agreement_mode(
        self,
    ):
        jwe = JsonWebEncryption()
        bob_key = OKPKey.generate_key("X25519", is_private=True)
        charlie_key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-ES", "enc": "A128GCM"}
        header_obj = {"protected": protected}
        self.assertRaises(
            InvalidAlgorithmForMultipleRecipientsMode,
            jwe.serialize_json,
            header_obj,
            b"hello",
            [bob_key, charlie_key],
        )

    def test_ecdh_es_decryption_with_public_key_fails(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-ES", "enc": "A128GCM"}

        key = {
            "kty": "EC",
            "crv": "P-256",
            "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
            "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        }
        data = jwe.serialize_compact(protected, b"hello", key)
        self.assertRaises(ValueError, jwe.deserialize_compact, data, key)

    def test_ecdh_es_encryption_fails_if_key_curve_is_inappropriate(self):
        jwe = JsonWebEncryption()
        protected = {"alg": "ECDH-ES", "enc": "A128GCM"}

        key = OKPKey.generate_key("Ed25519", is_private=False)
        self.assertRaises(ValueError, jwe.serialize_compact, protected, b"hello", key)

    def test_ecdh_es_decryption_fails_if_key_matches_to_no_recipient(self):
        jwe = JsonWebEncryption()

        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        protected = {
            "alg": "ECDH-ES+A256KW",
            "enc": "A256GCM",
            "apu": "QWxpY2U",
            "apv": "Qm9i",
        }

        unprotected = {"jku": "https://alice.example.com/keys.jwks"}

        recipients = [{"header": {"kid": "bob-key-2"}}]

        jwe_aad = b"Authenticate me too."

        header_obj = {
            "protected": protected,
            "unprotected": unprotected,
            "recipients": recipients,
            "aad": jwe_aad,
        }

        payload = b"Three is a magic number."

        data = jwe.serialize_json(header_obj, payload, bob_key)

        self.assertRaises(InvalidUnwrap, jwe.deserialize_json, data, charlie_key)

    def test_decryption_with_json_serialization_succeeds_while_encrypted_key_for_another_recipient_is_invalid(
        self,
    ):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kid": "Alice's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        OKPKey.import_key(
            {
                "kid": "Bob's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kid": "Charlie's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        data = {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
            + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
            + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
            + "RnFVQUZhMzlkeUJjIn19",
            "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"kid": "Bob's key"},
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                    + "eU1cSl55cQ0hGezJu2N9IY0QM",  # Invalid encrypted key
                },
                {
                    "header": {"kid": "Charlie's key"},
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                    + "fe4z3PQ2YH2afvjQ28aiCTWFE",  # Valid encrypted key
                },
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
        }

        rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

        self.assertEqual(rv_at_charlie.keys(), {"header", "payload"})

        self.assertEqual(
            rv_at_charlie["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv_at_charlie["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv_at_charlie["header"]["unprotected"],
            {"jku": "https://alice.example.com/keys.jwks"},
        )

        self.assertEqual(
            rv_at_charlie["header"]["recipients"],
            [{"header": {"kid": "Bob's key"}}, {"header": {"kid": "Charlie's key"}}],
        )

        self.assertEqual(rv_at_charlie["payload"], b"Three is a magic number.")

    def test_decryption_with_json_serialization_fails_if_encrypted_key_for_this_recipient_is_invalid(
        self,
    ):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kid": "Alice's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kid": "Bob's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        OKPKey.import_key(
            {
                "kid": "Charlie's key",
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        data = {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
            + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
            + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
            + "RnFVQUZhMzlkeUJjIn19",
            "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"kid": "Bob's key"},
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                    + "eU1cSl55cQ0hGezJu2N9IY0QM",  # Invalid encrypted key
                },
                {
                    "header": {"kid": "Charlie's key"},
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                    + "fe4z3PQ2YH2afvjQ28aiCTWFE",  # Valid encrypted key
                },
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
        }

        self.assertRaises(
            InvalidUnwrap, jwe.deserialize_json, data, bob_key, sender_key=alice_key
        )

    def test_dir_alg(self):
        jwe = JsonWebEncryption()
        key = OctKey.generate_key(128, is_private=True)
        protected = {"alg": "dir", "enc": "A128GCM"}
        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        self.assertEqual(rv["payload"], b"hello")

        key2 = OctKey.generate_key(256, is_private=True)
        self.assertRaises(ValueError, jwe.deserialize_compact, data, key2)

        self.assertRaises(ValueError, jwe.serialize_compact, protected, b"hello", key2)

    def test_decryption_of_message_to_multiple_recipients_by_matching_key(self):
        jwe = JsonWebEncryption()

        alice_public_key = OKPKey.import_key(
            {
                "kid": "WjKgJV7VRw3hmgU6--4v15c0Aewbcvat1BsRFTIqa5Q",
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            }
        )

        key_store = {}

        charlie_X448_key_id = (
            "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
        )
        charlie_X448_key = OKPKey.import_key(
            {
                "kid": "_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw",
                "kty": "OKP",
                "crv": "X448",
                "x": "M-OMugy74ksznVQ-Bp6MC_-GEPSrT8yiAtminJvw0j_UxJtpNHl_hcWMSf_Pfm_ws0vVWvAfwwA",
                "d": "VGZPkclj_7WbRaRMzBqxpzXIpc2xz1d3N1ay36UxdVLfKaP33hABBMpddTRv1f-hRsQUNvmlGOg",
            }
        )
        key_store[charlie_X448_key_id] = charlie_X448_key

        charlie_X25519_key_id = (
            "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
        )
        charlie_X25519_key = OKPKey.import_key(
            {
                "kid": "ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec",
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )
        key_store[charlie_X25519_key_id] = charlie_X25519_key

        data = """
            {
                "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
                "unprotected": {
                    "jku": "https://alice.example.com/keys.jwks"
                },
                "recipients": [
                    {
                        "header": {
                            "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
                        },
                        "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                    },
                    {
                        "header": {
                            "kid": "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
                        },
                        "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                    }
                ],
                "iv": "AAECAwQFBgcICQoLDA0ODw",
                "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
                "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
            }"""

        parsed_data = jwe.parse_json(data)

        available_key_id = next(
            recipient["header"]["kid"]
            for recipient in parsed_data["recipients"]
            if recipient["header"]["kid"] in key_store.keys()
        )
        available_key = key_store[available_key_id]

        rv = jwe.deserialize_json(
            parsed_data, (available_key_id, available_key), sender_key=alice_public_key
        )

        self.assertEqual(rv.keys(), {"header", "payload"})

        self.assertEqual(
            rv["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv["header"]["unprotected"], {"jku": "https://alice.example.com/keys.jwks"}
        )

        self.assertEqual(
            rv["header"]["recipients"],
            [
                {
                    "header": {
                        "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
                    }
                },
                {
                    "header": {
                        "kid": "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
                    }
                },
            ],
        )

        self.assertEqual(rv["payload"], b"Three is a magic number.")

    def test_decryption_of_json_string(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )
        charlie_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
                "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
            }
        )

        data = """
            {
                "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
                "unprotected": {
                    "jku": "https://alice.example.com/keys.jwks"
                },
                "recipients": [
                    {
                        "header": {
                            "kid": "bob-key-2"
                        },
                        "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                    },
                    {
                        "header": {
                            "kid": "2021-05-06"
                        },
                        "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                    }
                ],
                "iv": "AAECAwQFBgcICQoLDA0ODw",
                "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
                "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
            }"""

        rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

        self.assertEqual(rv_at_bob.keys(), {"header", "payload"})

        self.assertEqual(
            rv_at_bob["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv_at_bob["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv_at_bob["header"]["unprotected"],
            {"jku": "https://alice.example.com/keys.jwks"},
        )

        self.assertEqual(
            rv_at_bob["header"]["recipients"],
            [{"header": {"kid": "bob-key-2"}}, {"header": {"kid": "2021-05-06"}}],
        )

        self.assertEqual(rv_at_bob["payload"], b"Three is a magic number.")

        rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

        self.assertEqual(rv_at_charlie.keys(), {"header", "payload"})

        self.assertEqual(
            rv_at_charlie["header"].keys(), {"protected", "unprotected", "recipients"}
        )

        self.assertEqual(
            rv_at_charlie["header"]["protected"],
            {
                "alg": "ECDH-1PU+A128KW",
                "enc": "A256CBC-HS512",
                "apu": "QWxpY2U",
                "apv": "Qm9iIGFuZCBDaGFybGll",
                "epk": {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                },
            },
        )

        self.assertEqual(
            rv_at_charlie["header"]["unprotected"],
            {"jku": "https://alice.example.com/keys.jwks"},
        )

        self.assertEqual(
            rv_at_charlie["header"]["recipients"],
            [{"header": {"kid": "bob-key-2"}}, {"header": {"kid": "2021-05-06"}}],
        )

        self.assertEqual(rv_at_charlie["payload"], b"Three is a magic number.")

    def test_parse_json(self):
        json_msg = """
            {
                "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
                "unprotected": {
                    "jku": "https://alice.example.com/keys.jwks"
                },
                "recipients": [
                    {
                        "header": {
                            "kid": "bob-key-2"
                        },
                        "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                    },
                    {
                        "header": {
                            "kid": "2021-05-06"
                        },
                        "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                    }
                ],
                "iv": "AAECAwQFBgcICQoLDA0ODw",
                "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
                "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
            }"""

        parsed_msg = JsonWebEncryption.parse_json(json_msg)

        self.assertEqual(
            parsed_msg,
            {
                "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
                "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
                "recipients": [
                    {
                        "header": {"kid": "bob-key-2"},
                        "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN",
                    },
                    {
                        "header": {"kid": "2021-05-06"},
                        "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE",
                    },
                ],
                "iv": "AAECAwQFBgcICQoLDA0ODw",
                "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
                "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
            },
        )

    def test_parse_json_fails_if_json_msg_is_invalid(self):
        json_msg = """
            {
                "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
                "unprotected": {
                    "jku": "https://alice.example.com/keys.jwks"
                },
                "recipients": [
                    {
                        "header": {
                            "kid": "bob-key-2"
                        ,
                        "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                    },
                    {
                        "header": {
                            "kid": "2021-05-06"
                        },
                        "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                    }
                ],
                "iv": "AAECAwQFBgcICQoLDA0ODw",
                "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
                "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
            }"""

        self.assertRaises(DecodeError, JsonWebEncryption.parse_json, json_msg)

    def test_decryption_fails_if_ciphertext_is_invalid(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
                "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
            }
        )
        bob_key = OKPKey.import_key(
            {
                "kty": "OKP",
                "crv": "X25519",
                "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
                "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
            }
        )

        data = {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
            + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
            + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
            + "RnFVQUZhMzlkeUJjIn19",
            "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"kid": "bob-key-2"},
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                    + "eU1cSl55cQ0hGezJu2N9IY0QN",
                }
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFY",  # invalid ciphertext
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
        }

        self.assertRaises(
            Exception, jwe.deserialize_json, data, bob_key, sender_key=alice_key
        )

    def test_generic_serialize_deserialize_for_compact_serialization(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        header_obj = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}

        data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
        self.assertIsInstance(data, bytes)

        rv = jwe.deserialize(data, bob_key, sender_key=alice_key)
        self.assertEqual(rv["payload"], b"hello")

    def test_generic_serialize_deserialize_for_json_serialization(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
        self.assertIsInstance(data, dict)

        rv = jwe.deserialize(data, bob_key, sender_key=alice_key)
        self.assertEqual(rv["payload"], b"hello")

    def test_generic_deserialize_for_json_serialization_string(self):
        jwe = JsonWebEncryption()

        alice_key = OKPKey.generate_key("X25519", is_private=True)
        bob_key = OKPKey.generate_key("X25519", is_private=True)

        protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
        header_obj = {"protected": protected}

        data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
        self.assertIsInstance(data, dict)

        data_as_string = json.dumps(data)

        rv = jwe.deserialize(data_as_string, bob_key, sender_key=alice_key)
        self.assertEqual(rv["payload"], b"hello")



import time
from unittest import TestCase
from unittest import mock

from authlib.jose import jwt
from authlib.oauth2.rfc7523 import ClientSecretJWT
from authlib.oauth2.rfc7523 import PrivateKeyJWT
from tests.util import read_file_path


class ClientSecretJWTTest(TestCase):
    def test_nothing_set(self):
        jwt_signer = ClientSecretJWT()

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "HS256")

    def test_endpoint_set(self):
        jwt_signer = ClientSecretJWT(
            token_endpoint="https://example.com/oauth/access_token"
        )

        self.assertEqual(
            jwt_signer.token_endpoint, "https://example.com/oauth/access_token"
        )
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "HS256")

    def test_alg_set(self):
        jwt_signer = ClientSecretJWT(alg="HS512")

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "HS512")

    def test_claims_set(self):
        jwt_signer = ClientSecretJWT(claims={"foo1": "bar1"})

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, {"foo1": "bar1"})
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "HS256")

    def test_headers_set(self):
        jwt_signer = ClientSecretJWT(headers={"foo1": "bar1"})

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, {"foo1": "bar1"})
        self.assertEqual(jwt_signer.alg, "HS256")

    def test_all_set(self):
        jwt_signer = ClientSecretJWT(
            token_endpoint="https://example.com/oauth/access_token",
            claims={"foo1a": "bar1a"},
            headers={"foo1b": "bar1b"},
            alg="HS512",
        )

        self.assertEqual(
            jwt_signer.token_endpoint, "https://example.com/oauth/access_token"
        )
        self.assertEqual(jwt_signer.claims, {"foo1a": "bar1a"})
        self.assertEqual(jwt_signer.headers, {"foo1b": "bar1b"})
        self.assertEqual(jwt_signer.alg, "HS512")

    @staticmethod
    def sign_and_decode(jwt_signer, client_id, client_secret, token_endpoint):
        auth = mock.MagicMock()
        auth.client_id = client_id
        auth.client_secret = client_secret

        pre_sign_time = int(time.time())

        data = jwt_signer.sign(auth, token_endpoint).decode("utf-8")
        decoded = jwt.decode(
            data, client_secret
        )  # , claims_cls=None, claims_options=None, claims_params=None):

        iat = decoded.pop("iat")
        exp = decoded.pop("exp")
        jti = decoded.pop("jti")

        return decoded, pre_sign_time, iat, exp, jti

    def test_sign_nothing_set(self):
        jwt_signer = ClientSecretJWT()

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
            decoded,
        )

        self.assertEqual({"alg": "HS256", "typ": "JWT"}, decoded.header)

    def test_sign_custom_jti(self):
        jwt_signer = ClientSecretJWT(claims={"jti": "custom_jti"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertEqual("custom_jti", jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual({"alg": "HS256", "typ": "JWT"}, decoded.header)

    def test_sign_with_additional_header(self):
        jwt_signer = ClientSecretJWT(headers={"kid": "custom_kid"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual(
            {"alg": "HS256", "typ": "JWT", "kid": "custom_kid"}, decoded.header
        )

    def test_sign_with_additional_headers(self):
        jwt_signer = ClientSecretJWT(
            headers={"kid": "custom_kid", "jku": "https://example.com/oauth/jwks"}
        )

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual(
            {
                "alg": "HS256",
                "typ": "JWT",
                "kid": "custom_kid",
                "jku": "https://example.com/oauth/jwks",
            },
            decoded.header,
        )

    def test_sign_with_additional_claim(self):
        jwt_signer = ClientSecretJWT(claims={"name": "Foo"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
                "name": "Foo",
            },
        )

        self.assertEqual({"alg": "HS256", "typ": "JWT"}, decoded.header)

    def test_sign_with_additional_claims(self):
        jwt_signer = ClientSecretJWT(claims={"name": "Foo", "role": "bar"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            "client_secret_1",
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
                "name": "Foo",
                "role": "bar",
            },
        )

        self.assertEqual({"alg": "HS256", "typ": "JWT"}, decoded.header)


class PrivateKeyJWTTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.public_key = read_file_path("rsa_public.pem")
        cls.private_key = read_file_path("rsa_private.pem")

    def test_nothing_set(self):
        jwt_signer = PrivateKeyJWT()

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "RS256")

    def test_endpoint_set(self):
        jwt_signer = PrivateKeyJWT(
            token_endpoint="https://example.com/oauth/access_token"
        )

        self.assertEqual(
            jwt_signer.token_endpoint, "https://example.com/oauth/access_token"
        )
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "RS256")

    def test_alg_set(self):
        jwt_signer = PrivateKeyJWT(alg="RS512")

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "RS512")

    def test_claims_set(self):
        jwt_signer = PrivateKeyJWT(claims={"foo1": "bar1"})

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, {"foo1": "bar1"})
        self.assertEqual(jwt_signer.headers, None)
        self.assertEqual(jwt_signer.alg, "RS256")

    def test_headers_set(self):
        jwt_signer = PrivateKeyJWT(headers={"foo1": "bar1"})

        self.assertEqual(jwt_signer.token_endpoint, None)
        self.assertEqual(jwt_signer.claims, None)
        self.assertEqual(jwt_signer.headers, {"foo1": "bar1"})
        self.assertEqual(jwt_signer.alg, "RS256")

    def test_all_set(self):
        jwt_signer = PrivateKeyJWT(
            token_endpoint="https://example.com/oauth/access_token",
            claims={"foo1a": "bar1a"},
            headers={"foo1b": "bar1b"},
            alg="RS512",
        )

        self.assertEqual(
            jwt_signer.token_endpoint, "https://example.com/oauth/access_token"
        )
        self.assertEqual(jwt_signer.claims, {"foo1a": "bar1a"})
        self.assertEqual(jwt_signer.headers, {"foo1b": "bar1b"})
        self.assertEqual(jwt_signer.alg, "RS512")

    @staticmethod
    def sign_and_decode(jwt_signer, client_id, public_key, private_key, token_endpoint):
        auth = mock.MagicMock()
        auth.client_id = client_id
        auth.client_secret = private_key

        pre_sign_time = int(time.time())

        data = jwt_signer.sign(auth, token_endpoint).decode("utf-8")
        decoded = jwt.decode(
            data, public_key
        )  # , claims_cls=None, claims_options=None, claims_params=None):

        iat = decoded.pop("iat")
        exp = decoded.pop("exp")
        jti = decoded.pop("jti")

        return decoded, pre_sign_time, iat, exp, jti

    def test_sign_nothing_set(self):
        jwt_signer = PrivateKeyJWT()

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
            decoded,
        )

        self.assertEqual({"alg": "RS256", "typ": "JWT"}, decoded.header)

    def test_sign_custom_jti(self):
        jwt_signer = PrivateKeyJWT(claims={"jti": "custom_jti"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertEqual("custom_jti", jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual({"alg": "RS256", "typ": "JWT"}, decoded.header)

    def test_sign_with_additional_header(self):
        jwt_signer = PrivateKeyJWT(headers={"kid": "custom_kid"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual(
            {"alg": "RS256", "typ": "JWT", "kid": "custom_kid"}, decoded.header
        )

    def test_sign_with_additional_headers(self):
        jwt_signer = PrivateKeyJWT(
            headers={"kid": "custom_kid", "jku": "https://example.com/oauth/jwks"}
        )

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
            },
        )

        self.assertEqual(
            {
                "alg": "RS256",
                "typ": "JWT",
                "kid": "custom_kid",
                "jku": "https://example.com/oauth/jwks",
            },
            decoded.header,
        )

    def test_sign_with_additional_claim(self):
        jwt_signer = PrivateKeyJWT(claims={"name": "Foo"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
                "name": "Foo",
            },
        )

        self.assertEqual({"alg": "RS256", "typ": "JWT"}, decoded.header)

    def test_sign_with_additional_claims(self):
        jwt_signer = PrivateKeyJWT(claims={"name": "Foo", "role": "bar"})

        decoded, pre_sign_time, iat, exp, jti = self.sign_and_decode(
            jwt_signer,
            "client_id_1",
            self.public_key,
            self.private_key,
            "https://example.com/oauth/access_token",
        )

        self.assertGreaterEqual(iat, pre_sign_time)
        self.assertGreaterEqual(exp, iat + 3600)
        self.assertLessEqual(exp, iat + 3600 + 2)
        self.assertIsNotNone(jti)

        self.assertEqual(
            decoded,
            {
                "iss": "client_id_1",
                "aud": "https://example.com/oauth/access_token",
                "sub": "client_id_1",
                "name": "Foo",
                "role": "bar",
            },
        )

        self.assertEqual({"alg": "RS256", "typ": "JWT"}, decoded.header)



import time
from copy import deepcopy
from unittest import TestCase
from unittest import mock

from authlib.common.security import generate_token
from authlib.common.urls import add_params_to_uri
from authlib.common.urls import url_encode
from authlib.integrations.requests_client import OAuth2Session
from authlib.integrations.requests_client import OAuthError
from authlib.oauth2.rfc6749 import MismatchingStateException
from authlib.oauth2.rfc7523 import ClientSecretJWT
from authlib.oauth2.rfc7523 import PrivateKeyJWT

from ..util import read_key_file


def mock_json_response(payload):
    def fake_send(r, **kwargs):
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: payload
        return resp

    return fake_send


def mock_assertion_response(ctx, session):
    def fake_send(r, **kwargs):
        ctx.assertIn("client_assertion=", r.body)
        ctx.assertIn("client_assertion_type=", r.body)
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: ctx.token
        return resp

    session.send = fake_send


class OAuth2SessionTest(TestCase):
    def setUp(self):
        self.token = {
            "token_type": "Bearer",
            "access_token": "a",
            "refresh_token": "b",
            "expires_in": "3600",
            "expires_at": int(time.time()) + 3600,
        }
        self.client_id = "foo"

    def test_invalid_token_type(self):
        token = {
            "token_type": "invalid",
            "access_token": "a",
            "refresh_token": "b",
            "expires_in": "3600",
            "expires_at": int(time.time()) + 3600,
        }
        with OAuth2Session(self.client_id, token=token) as sess:
            self.assertRaises(OAuthError, sess.get, "https://i.b")

    def test_add_token_to_header(self):
        token = "Bearer " + self.token["access_token"]

        def verifier(r, **kwargs):
            auth_header = r.headers.get("Authorization", None)
            self.assertEqual(auth_header, token)
            resp = mock.MagicMock()
            return resp

        sess = OAuth2Session(client_id=self.client_id, token=self.token)
        sess.send = verifier
        sess.get("https://i.b")

    def test_add_token_to_body(self):
        def verifier(r, **kwargs):
            self.assertIn(self.token["access_token"], r.body)
            resp = mock.MagicMock()
            return resp

        sess = OAuth2Session(
            client_id=self.client_id, token=self.token, token_placement="body"
        )
        sess.send = verifier
        sess.post("https://i.b")

    def test_add_token_to_uri(self):
        def verifier(r, **kwargs):
            self.assertIn(self.token["access_token"], r.url)
            resp = mock.MagicMock()
            return resp

        sess = OAuth2Session(
            client_id=self.client_id, token=self.token, token_placement="uri"
        )
        sess.send = verifier
        sess.get("https://i.b")

    def test_create_authorization_url(self):
        url = "https://example.com/authorize?foo=bar"

        sess = OAuth2Session(client_id=self.client_id)
        auth_url, state = sess.create_authorization_url(url)
        self.assertIn(state, auth_url)
        self.assertIn(self.client_id, auth_url)
        self.assertIn("response_type=code", auth_url)

        sess = OAuth2Session(client_id=self.client_id, prompt="none")
        auth_url, state = sess.create_authorization_url(
            url, state="foo", redirect_uri="https://i.b", scope="profile"
        )
        self.assertEqual(state, "foo")
        self.assertIn("i.b", auth_url)
        self.assertIn("profile", auth_url)
        self.assertIn("prompt=none", auth_url)

    def test_code_challenge(self):
        sess = OAuth2Session(client_id=self.client_id, code_challenge_method="S256")

        url = "https://example.com/authorize"
        auth_url, _ = sess.create_authorization_url(
            url, code_verifier=generate_token(48)
        )
        self.assertIn("code_challenge", auth_url)
        self.assertIn("code_challenge_method=S256", auth_url)

    def test_token_from_fragment(self):
        sess = OAuth2Session(self.client_id)
        response_url = "https://i.b/callback#" + url_encode(self.token.items())
        self.assertEqual(sess.token_from_fragment(response_url), self.token)
        token = sess.fetch_token(authorization_response=response_url)
        self.assertEqual(token, self.token)

    def test_fetch_token_post(self):
        url = "https://example.com/token"

        def fake_send(r, **kwargs):
            self.assertIn("code=v", r.body)
            self.assertIn("client_id=", r.body)
            self.assertIn("grant_type=authorization_code", r.body)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess = OAuth2Session(client_id=self.client_id)
        sess.send = fake_send
        self.assertEqual(
            sess.fetch_token(url, authorization_response="https://i.b/?code=v"),
            self.token,
        )

        sess = OAuth2Session(
            client_id=self.client_id,
            token_endpoint_auth_method="none",
        )
        sess.send = fake_send
        token = sess.fetch_token(url, code="v")
        self.assertEqual(token, self.token)

        error = {"error": "invalid_request"}
        sess = OAuth2Session(client_id=self.client_id, token=self.token)
        sess.send = mock_json_response(error)
        self.assertRaises(OAuthError, sess.fetch_access_token, url)

    def test_fetch_token_get(self):
        url = "https://example.com/token"

        def fake_send(r, **kwargs):
            self.assertIn("code=v", r.url)
            self.assertIn("grant_type=authorization_code", r.url)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess = OAuth2Session(client_id=self.client_id)
        sess.send = fake_send
        token = sess.fetch_token(
            url, authorization_response="https://i.b/?code=v", method="GET"
        )
        self.assertEqual(token, self.token)

        sess = OAuth2Session(
            client_id=self.client_id,
            token_endpoint_auth_method="none",
        )
        sess.send = fake_send
        token = sess.fetch_token(url, code="v", method="GET")
        self.assertEqual(token, self.token)

        token = sess.fetch_token(url + "?q=a", code="v", method="GET")
        self.assertEqual(token, self.token)

    def test_token_auth_method_client_secret_post(self):
        url = "https://example.com/token"

        def fake_send(r, **kwargs):
            self.assertIn("code=v", r.body)
            self.assertIn("client_id=", r.body)
            self.assertIn("client_secret=bar", r.body)
            self.assertIn("grant_type=authorization_code", r.body)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess = OAuth2Session(
            client_id=self.client_id,
            client_secret="bar",
            token_endpoint_auth_method="client_secret_post",
        )
        sess.send = fake_send
        token = sess.fetch_token(url, code="v")
        self.assertEqual(token, self.token)

    def test_access_token_response_hook(self):
        url = "https://example.com/token"

        def access_token_response_hook(resp):
            self.assertEqual(resp.json(), self.token)
            return resp

        sess = OAuth2Session(client_id=self.client_id, token=self.token)
        sess.register_compliance_hook(
            "access_token_response", access_token_response_hook
        )
        sess.send = mock_json_response(self.token)
        self.assertEqual(sess.fetch_token(url), self.token)

    def test_password_grant_type(self):
        url = "https://example.com/token"

        def fake_send(r, **kwargs):
            self.assertIn("username=v", r.body)
            self.assertIn("grant_type=password", r.body)
            self.assertIn("scope=profile", r.body)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess = OAuth2Session(client_id=self.client_id, scope="profile")
        sess.send = fake_send
        token = sess.fetch_token(url, username="v", password="v")
        self.assertEqual(token, self.token)

    def test_client_credentials_type(self):
        url = "https://example.com/token"

        def fake_send(r, **kwargs):
            self.assertIn("grant_type=client_credentials", r.body)
            self.assertIn("scope=profile", r.body)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess = OAuth2Session(
            client_id=self.client_id,
            client_secret="v",
            scope="profile",
        )
        sess.send = fake_send
        token = sess.fetch_token(url)
        self.assertEqual(token, self.token)

    def test_cleans_previous_token_before_fetching_new_one(self):
        """Makes sure the previous token is cleaned before fetching a new one.
        The reason behind it is that, if the previous token is expired, this
        method shouldn't fail with a TokenExpiredError, since it's attempting
        to get a new one (which shouldn't be expired).
        """
        now = int(time.time())
        new_token = deepcopy(self.token)
        past = now - 7200
        self.token["expires_at"] = past
        new_token["expires_at"] = now + 3600
        url = "https://example.com/token"

        with mock.patch("time.time", lambda: now):
            sess = OAuth2Session(client_id=self.client_id, token=self.token)
            sess.send = mock_json_response(new_token)
            self.assertEqual(sess.fetch_token(url), new_token)

    def test_mis_match_state(self):
        sess = OAuth2Session("foo")
        self.assertRaises(
            MismatchingStateException,
            sess.fetch_token,
            "https://i.b/token",
            authorization_response="https://i.b/no-state?code=abc",
            state="somestate",
        )

    def test_token_status(self):
        token = dict(access_token="a", token_type="bearer", expires_at=100)
        sess = OAuth2Session("foo", token=token)

        self.assertTrue(sess.token.is_expired)

    def test_token_status2(self):
        token = dict(access_token="a", token_type="bearer", expires_in=10)
        sess = OAuth2Session("foo", token=token, leeway=15)

        self.assertTrue(sess.token.is_expired(sess.leeway))

    def test_token_status3(self):
        token = dict(access_token="a", token_type="bearer", expires_in=10)
        sess = OAuth2Session("foo", token=token, leeway=5)

        self.assertFalse(sess.token.is_expired(sess.leeway))

    def test_token_expired(self):
        token = dict(access_token="a", token_type="bearer", expires_at=100)
        sess = OAuth2Session("foo", token=token)
        self.assertRaises(
            OAuthError,
            sess.get,
            "https://i.b/token",
        )

    def test_missing_token(self):
        sess = OAuth2Session("foo")
        self.assertRaises(
            OAuthError,
            sess.get,
            "https://i.b/token",
        )

    def test_register_compliance_hook(self):
        sess = OAuth2Session("foo")
        self.assertRaises(
            ValueError,
            sess.register_compliance_hook,
            "invalid_hook",
            lambda o: o,
        )

        def protected_request(url, headers, data):
            self.assertIn("Authorization", headers)
            return url, headers, data

        sess = OAuth2Session("foo", token=self.token)
        sess.register_compliance_hook(
            "protected_request",
            protected_request,
        )
        sess.send = mock_json_response({"name": "a"})
        sess.get("https://i.b/user")

    def test_auto_refresh_token(self):
        def _update_token(token, refresh_token=None, access_token=None):
            self.assertEqual(refresh_token, "b")
            self.assertEqual(token, self.token)

        update_token = mock.Mock(side_effect=_update_token)
        old_token = dict(
            access_token="a", refresh_token="b", token_type="bearer", expires_at=100
        )
        sess = OAuth2Session(
            "foo",
            token=old_token,
            token_endpoint="https://i.b/token",
            update_token=update_token,
        )
        sess.send = mock_json_response(self.token)
        sess.get("https://i.b/user")
        self.assertTrue(update_token.called)

    def test_auto_refresh_token2(self):
        def _update_token(token, refresh_token=None, access_token=None):
            self.assertEqual(access_token, "a")
            self.assertEqual(token, self.token)

        update_token = mock.Mock(side_effect=_update_token)
        old_token = dict(access_token="a", token_type="bearer", expires_at=100)

        sess = OAuth2Session(
            "foo",
            token=old_token,
            token_endpoint="https://i.b/token",
            grant_type="client_credentials",
        )
        sess.send = mock_json_response(self.token)
        sess.get("https://i.b/user")
        self.assertFalse(update_token.called)

        sess = OAuth2Session(
            "foo",
            token=old_token,
            token_endpoint="https://i.b/token",
            grant_type="client_credentials",
            update_token=update_token,
        )
        sess.send = mock_json_response(self.token)
        sess.get("https://i.b/user")
        self.assertTrue(update_token.called)

    def test_revoke_token(self):
        sess = OAuth2Session("a")
        answer = {"status": "ok"}
        sess.send = mock_json_response(answer)
        resp = sess.revoke_token("https://i.b/token", "hi")
        self.assertEqual(resp.json(), answer)
        resp = sess.revoke_token(
            "https://i.b/token", "hi", token_type_hint="access_token"
        )
        self.assertEqual(resp.json(), answer)

        def revoke_token_request(url, headers, data):
            self.assertEqual(url, "https://i.b/token")
            return url, headers, data

        sess.register_compliance_hook(
            "revoke_token_request",
            revoke_token_request,
        )
        sess.revoke_token(
            "https://i.b/token", "hi", body="", token_type_hint="access_token"
        )

    def test_introspect_token(self):
        sess = OAuth2Session("a")
        answer = {
            "active": True,
            "client_id": "l238j323ds-23ij4",
            "username": "jdoe",
            "scope": "read write dolphin",
            "sub": "Z5O3upPC88QrAjx00dis",
            "aud": "https://protected.example.net/resource",
            "iss": "https://server.example.com/",
            "exp": 1419356238,
            "iat": 1419350238,
        }
        sess.send = mock_json_response(answer)
        resp = sess.introspect_token("https://i.b/token", "hi")
        self.assertEqual(resp.json(), answer)

    def test_client_secret_jwt(self):
        sess = OAuth2Session(
            "id", "secret", token_endpoint_auth_method="client_secret_jwt"
        )
        sess.register_client_auth_method(ClientSecretJWT())

        mock_assertion_response(self, sess)
        token = sess.fetch_token("https://i.b/token")
        self.assertEqual(token, self.token)

    def test_client_secret_jwt2(self):
        sess = OAuth2Session(
            "id",
            "secret",
            token_endpoint_auth_method=ClientSecretJWT(),
        )
        mock_assertion_response(self, sess)
        token = sess.fetch_token("https://i.b/token")
        self.assertEqual(token, self.token)

    def test_private_key_jwt(self):
        client_secret = read_key_file("rsa_private.pem")
        sess = OAuth2Session(
            "id", client_secret, token_endpoint_auth_method="private_key_jwt"
        )
        sess.register_client_auth_method(PrivateKeyJWT())
        mock_assertion_response(self, sess)
        token = sess.fetch_token("https://i.b/token")
        self.assertEqual(token, self.token)

    def test_custom_client_auth_method(self):
        def auth_client(client, method, uri, headers, body):
            uri = add_params_to_uri(
                uri,
                [
                    ("client_id", client.client_id),
                    ("client_secret", client.client_secret),
                ],
            )
            uri = uri + "&" + body
            body = ""
            return uri, headers, body

        sess = OAuth2Session(
            "id", "secret", token_endpoint_auth_method="client_secret_uri"
        )
        sess.register_client_auth_method(("client_secret_uri", auth_client))

        def fake_send(r, **kwargs):
            self.assertIn("client_id=", r.url)
            self.assertIn("client_secret=", r.url)
            resp = mock.MagicMock()
            resp.status_code = 200
            resp.json = lambda: self.token
            return resp

        sess.send = fake_send
        token = sess.fetch_token("https://i.b/token")
        self.assertEqual(token, self.token)

    def test_use_client_token_auth(self):
        import requests

        token = "Bearer " + self.token["access_token"]

        def verifier(r, **kwargs):
            auth_header = r.headers.get("Authorization", None)
            self.assertEqual(auth_header, token)
            resp = mock.MagicMock()
            return resp

        client = OAuth2Session(client_id=self.client_id, token=self.token)

        sess = requests.Session()
        sess.send = verifier
        sess.get("https://i.b", auth=client.token_auth)

    def test_use_default_request_timeout(self):
        expected_timeout = 15

        def verifier(r, **kwargs):
            timeout = kwargs.get("timeout")
            self.assertEqual(timeout, expected_timeout)
            resp = mock.MagicMock()
            return resp

        client = OAuth2Session(
            client_id=self.client_id,
            token=self.token,
            default_timeout=expected_timeout,
        )

        client.send = verifier
        client.request("GET", "https://i.b", withhold_token=False)

    def test_override_default_request_timeout(self):
        default_timeout = 15
        expected_timeout = 10

        def verifier(r, **kwargs):
            timeout = kwargs.get("timeout")
            self.assertEqual(timeout, expected_timeout)
            resp = mock.MagicMock()
            return resp

        client = OAuth2Session(
            client_id=self.client_id,
            token=self.token,
            default_timeout=default_timeout,
        )

        client.send = verifier
        client.request(
            "GET", "https://i.b", withhold_token=False, timeout=expected_timeout
        )



from unittest import TestCase
from unittest import mock

from cachelib import SimpleCache
from flask import Flask
from flask import session

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.flask_client import FlaskOAuth2App
from authlib.integrations.flask_client import OAuth
from authlib.integrations.flask_client import OAuthError
from authlib.jose import jwk
from authlib.oidc.core.grants.util import generate_id_token

from ..util import get_bearer_token
from ..util import mock_send_value


class FlaskOAuthTest(TestCase):
    def test_register_remote_app(self):
        app = Flask(__name__)
        oauth = OAuth(app)
        self.assertRaises(AttributeError, lambda: oauth.dev)

        oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
        )
        self.assertEqual(oauth.dev.name, "dev")
        self.assertEqual(oauth.dev.client_id, "dev")

    def test_register_conf_from_app(self):
        app = Flask(__name__)
        app.config.update(
            {
                "DEV_CLIENT_ID": "dev",
                "DEV_CLIENT_SECRET": "dev",
            }
        )
        oauth = OAuth(app)
        oauth.register("dev")
        self.assertEqual(oauth.dev.client_id, "dev")

    def test_register_with_overwrite(self):
        app = Flask(__name__)
        app.config.update(
            {
                "DEV_CLIENT_ID": "dev-1",
                "DEV_CLIENT_SECRET": "dev",
                "DEV_ACCESS_TOKEN_PARAMS": {"foo": "foo-1"},
            }
        )
        oauth = OAuth(app)
        oauth.register(
            "dev", overwrite=True, client_id="dev", access_token_params={"foo": "foo"}
        )
        self.assertEqual(oauth.dev.client_id, "dev-1")
        self.assertEqual(oauth.dev.client_secret, "dev")
        self.assertEqual(oauth.dev.access_token_params["foo"], "foo-1")

    def test_init_app_later(self):
        app = Flask(__name__)
        app.config.update(
            {
                "DEV_CLIENT_ID": "dev",
                "DEV_CLIENT_SECRET": "dev",
            }
        )
        oauth = OAuth()
        remote = oauth.register("dev")
        self.assertRaises(RuntimeError, lambda: oauth.dev.client_id)
        oauth.init_app(app)
        self.assertEqual(oauth.dev.client_id, "dev")
        self.assertEqual(remote.client_id, "dev")

        self.assertIsNone(oauth.cache)
        self.assertIsNone(oauth.fetch_token)
        self.assertIsNone(oauth.update_token)

    def test_init_app_params(self):
        app = Flask(__name__)
        oauth = OAuth()
        oauth.init_app(app, SimpleCache())
        self.assertIsNotNone(oauth.cache)
        self.assertIsNone(oauth.update_token)

        oauth.init_app(app, update_token=lambda o: o)
        self.assertIsNotNone(oauth.update_token)

    def test_create_client(self):
        app = Flask(__name__)
        oauth = OAuth(app)
        self.assertIsNone(oauth.create_client("dev"))
        oauth.register("dev", client_id="dev")
        self.assertIsNotNone(oauth.create_client("dev"))

    def test_register_oauth1_remote_app(self):
        app = Flask(__name__)
        oauth = OAuth(app)
        client_kwargs = dict(
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            fetch_request_token=lambda: None,
            save_request_token=lambda token: token,
        )
        oauth.register("dev", **client_kwargs)
        self.assertEqual(oauth.dev.name, "dev")
        self.assertEqual(oauth.dev.client_id, "dev")

        oauth = OAuth(app, cache=SimpleCache())
        oauth.register("dev", **client_kwargs)
        self.assertEqual(oauth.dev.name, "dev")
        self.assertEqual(oauth.dev.client_id, "dev")

    def test_oauth1_authorize_cache(self):
        app = Flask(__name__)
        app.secret_key = "!"
        cache = SimpleCache()
        oauth = OAuth(app, cache=cache)

        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(
                    "oauth_token=foo&oauth_verifier=baz"
                )
                resp = client.authorize_redirect("https://b.com/bar")
                self.assertEqual(resp.status_code, 302)
                url = resp.headers.get("Location")
                self.assertIn("oauth_token=foo", url)

        with app.test_request_context("/?oauth_token=foo"):
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(
                    "oauth_token=a&oauth_token_secret=b"
                )
                token = client.authorize_access_token()
                self.assertEqual(token["oauth_token"], "a")

    def test_oauth1_authorize_session(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(
                    "oauth_token=foo&oauth_verifier=baz"
                )
                resp = client.authorize_redirect("https://b.com/bar")
                self.assertEqual(resp.status_code, 302)
                url = resp.headers.get("Location")
                self.assertIn("oauth_token=foo", url)
                data = session["_state_dev_foo"]

        with app.test_request_context("/?oauth_token=foo"):
            session["_state_dev_foo"] = data
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(
                    "oauth_token=a&oauth_token_secret=b"
                )
                token = client.authorize_access_token()
                self.assertEqual(token["oauth_token"], "a")

    def test_register_oauth2_remote_app(self):
        app = Flask(__name__)
        oauth = OAuth(app)
        oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            refresh_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            update_token=lambda name: "hi",
        )
        self.assertEqual(oauth.dev.name, "dev")
        session = oauth.dev._get_oauth_client()
        self.assertIsNotNone(session.update_token)

    def test_oauth2_authorize(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with app.test_request_context():
            resp = client.authorize_redirect("https://b.com/bar")
            self.assertEqual(resp.status_code, 302)
            url = resp.headers.get("Location")
            self.assertIn("state=", url)
            state = dict(url_decode(urlparse.urlparse(url).query))["state"]
            self.assertIsNotNone(state)
            data = session[f"_state_dev_{state}"]

        with app.test_request_context(path=f"/?code=a&state={state}"):
            # session is cleared in tests
            session[f"_state_dev_{state}"] = data

            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(get_bearer_token())
                token = client.authorize_access_token()
                self.assertEqual(token["access_token"], "a")

        with app.test_request_context():
            self.assertEqual(client.token, None)

    def test_oauth2_authorize_access_denied(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with app.test_request_context(
            path="/?error=access_denied&error_description=Not+Allowed"
        ):
            # session is cleared in tests
            with mock.patch("requests.sessions.Session.send"):
                self.assertRaises(OAuthError, client.authorize_access_token)

    def test_oauth2_authorize_via_custom_client(self):
        class CustomRemoteApp(FlaskOAuth2App):
            OAUTH_APP_CONFIG = {"authorize_url": "https://i.b/custom"}

        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            client_cls=CustomRemoteApp,
        )
        with app.test_request_context():
            resp = client.authorize_redirect("https://b.com/bar")
            self.assertEqual(resp.status_code, 302)
            url = resp.headers.get("Location")
            self.assertTrue(url.startswith("https://i.b/custom?"))

    def test_oauth2_authorize_with_metadata(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
        )
        self.assertRaises(RuntimeError, lambda: client.create_authorization_url(None))

        client = oauth.register(
            "dev2",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            server_metadata_url="https://i.b/.well-known/openid-configuration",
        )
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(
                {"authorization_endpoint": "https://i.b/authorize"}
            )

            with app.test_request_context():
                resp = client.authorize_redirect("https://b.com/bar")
                self.assertEqual(resp.status_code, 302)

    def test_oauth2_authorize_code_challenge(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            client_kwargs={"code_challenge_method": "S256"},
        )

        with app.test_request_context():
            resp = client.authorize_redirect("https://b.com/bar")
            self.assertEqual(resp.status_code, 302)
            url = resp.headers.get("Location")
            self.assertIn("code_challenge=", url)
            self.assertIn("code_challenge_method=S256", url)

            state = dict(url_decode(urlparse.urlparse(url).query))["state"]
            self.assertIsNotNone(state)
            data = session[f"_state_dev_{state}"]

            verifier = data["data"]["code_verifier"]
            self.assertIsNotNone(verifier)

        def fake_send(sess, req, **kwargs):
            self.assertIn(f"code_verifier={verifier}", req.body)
            return mock_send_value(get_bearer_token())

        path = f"/?code=a&state={state}"
        with app.test_request_context(path=path):
            # session is cleared in tests
            session[f"_state_dev_{state}"] = data

            with mock.patch("requests.sessions.Session.send", fake_send):
                token = client.authorize_access_token()
                self.assertEqual(token["access_token"], "a")

    def test_openid_authorize(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        key = jwk.dumps("secret", "oct", kid="f")

        client = oauth.register(
            "dev",
            client_id="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            client_kwargs={"scope": "openid profile"},
            jwks={"keys": [key]},
        )

        with app.test_request_context():
            resp = client.authorize_redirect("https://b.com/bar")
            self.assertEqual(resp.status_code, 302)

            url = resp.headers["Location"]
            query_data = dict(url_decode(urlparse.urlparse(url).query))

            state = query_data["state"]
            self.assertIsNotNone(state)
            session_data = session[f"_state_dev_{state}"]
            nonce = session_data["data"]["nonce"]
            self.assertIsNotNone(nonce)
            self.assertEqual(nonce, query_data["nonce"])

        token = get_bearer_token()
        token["id_token"] = generate_id_token(
            token,
            {"sub": "123"},
            key,
            alg="HS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
            nonce=query_data["nonce"],
        )
        path = f"/?code=a&state={state}"
        with app.test_request_context(path=path):
            session[f"_state_dev_{state}"] = session_data
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(token)
                token = client.authorize_access_token()
                self.assertEqual(token["access_token"], "a")
                self.assertIn("userinfo", token)

    def test_oauth2_access_token_with_post(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )
        payload = {"code": "a", "state": "b"}
        with app.test_request_context(data=payload, method="POST"):
            session["_state_dev_b"] = {"data": payload}
            with mock.patch("requests.sessions.Session.send") as send:
                send.return_value = mock_send_value(get_bearer_token())
                token = client.authorize_access_token()
                self.assertEqual(token["access_token"], "a")

    def test_access_token_with_fetch_token(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth()

        token = get_bearer_token()
        oauth.init_app(app, fetch_token=lambda name: token)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        def fake_send(sess, req, **kwargs):
            auth = req.headers["Authorization"]
            self.assertEqual(auth, "Bearer {}".format(token["access_token"]))
            resp = mock.MagicMock()
            resp.text = "hi"
            resp.status_code = 200
            return resp

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send", fake_send):
                resp = client.get("/api/user")
                self.assertEqual(resp.text, "hi")

                # trigger ctx.authlib_client_oauth_token
                resp = client.get("/api/user")
                self.assertEqual(resp.text, "hi")

    def test_request_with_refresh_token(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth()

        expired_token = {
            "token_type": "Bearer",
            "access_token": "expired-a",
            "refresh_token": "expired-b",
            "expires_in": "3600",
            "expires_at": 1566465749,
        }
        oauth.init_app(app, fetch_token=lambda name: expired_token)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            refresh_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        def fake_send(sess, req, **kwargs):
            if req.url == "https://i.b/token":
                auth = req.headers["Authorization"]
                self.assertIn("Basic", auth)
                resp = mock.MagicMock()
                resp.json = get_bearer_token
                resp.status_code = 200
                return resp

            resp = mock.MagicMock()
            resp.text = "hi"
            resp.status_code = 200
            return resp

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send", fake_send):
                resp = client.get("/api/user", token=expired_token)
                self.assertEqual(resp.text, "hi")

    def test_request_without_token(self):
        app = Flask(__name__)
        app.secret_key = "!"
        oauth = OAuth(app)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        def fake_send(sess, req, **kwargs):
            auth = req.headers.get("Authorization")
            self.assertIsNone(auth)
            resp = mock.MagicMock()
            resp.text = "hi"
            resp.status_code = 200
            return resp

        with app.test_request_context():
            with mock.patch("requests.sessions.Session.send", fake_send):
                resp = client.get("/api/user", withhold_token=True)
                self.assertEqual(resp.text, "hi")
                self.assertRaises(OAuthError, client.get, "https://i.b/api/user")



from unittest import mock

from django.test import override_settings

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.django_client import OAuth
from authlib.integrations.django_client import OAuthError
from authlib.jose import JsonWebKey
from authlib.oidc.core.grants.util import generate_id_token
from tests.django_helper import TestCase

from ..util import get_bearer_token
from ..util import mock_send_value

dev_client = {"client_id": "dev-key", "client_secret": "dev-secret"}


class DjangoOAuthTest(TestCase):
    def test_register_remote_app(self):
        oauth = OAuth()
        self.assertRaises(AttributeError, lambda: oauth.dev)

        oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )
        self.assertEqual(oauth.dev.name, "dev")
        self.assertEqual(oauth.dev.client_id, "dev")

    def test_register_with_overwrite(self):
        oauth = OAuth()
        oauth.register(
            "dev_overwrite",
            overwrite=True,
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            access_token_params={"foo": "foo"},
            authorize_url="https://i.b/authorize",
        )
        self.assertEqual(oauth.dev_overwrite.client_id, "dev-client-id")
        self.assertEqual(oauth.dev_overwrite.access_token_params["foo"], "foo-1")

    @override_settings(AUTHLIB_OAUTH_CLIENTS={"dev": dev_client})
    def test_register_from_settings(self):
        oauth = OAuth()
        oauth.register("dev")
        self.assertEqual(oauth.dev.client_id, "dev-key")
        self.assertEqual(oauth.dev.client_secret, "dev-secret")

    def test_oauth1_authorize(self):
        request = self.factory.get("/login")
        request.session = self.factory.session

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            request_token_url="https://i.b/reqeust-token",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value("oauth_token=foo&oauth_verifier=baz")

            resp = client.authorize_redirect(request)
            self.assertEqual(resp.status_code, 302)
            url = resp.get("Location")
            self.assertIn("oauth_token=foo", url)

        request2 = self.factory.get(url)
        request2.session = request.session
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value("oauth_token=a&oauth_token_secret=b")
            token = client.authorize_access_token(request2)
            self.assertEqual(token["oauth_token"], "a")

    def test_oauth2_authorize(self):
        request = self.factory.get("/login")
        request.session = self.factory.session

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )
        rv = client.authorize_redirect(request, "https://a.b/c")
        self.assertEqual(rv.status_code, 302)
        url = rv.get("Location")
        self.assertIn("state=", url)
        state = dict(url_decode(urlparse.urlparse(url).query))["state"]

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())
            request2 = self.factory.get(f"/authorize?state={state}")
            request2.session = request.session

            token = client.authorize_access_token(request2)
            self.assertEqual(token["access_token"], "a")

    def test_oauth2_authorize_access_denied(self):
        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        with mock.patch("requests.sessions.Session.send"):
            request = self.factory.get(
                "/?error=access_denied&error_description=Not+Allowed"
            )
            request.session = self.factory.session
            self.assertRaises(OAuthError, client.authorize_access_token, request)

    def test_oauth2_authorize_code_challenge(self):
        request = self.factory.get("/login")
        request.session = self.factory.session

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            client_kwargs={"code_challenge_method": "S256"},
        )
        rv = client.authorize_redirect(request, "https://a.b/c")
        self.assertEqual(rv.status_code, 302)
        url = rv.get("Location")
        self.assertIn("state=", url)
        self.assertIn("code_challenge=", url)

        state = dict(url_decode(urlparse.urlparse(url).query))["state"]
        state_data = request.session[f"_state_dev_{state}"]["data"]
        verifier = state_data["code_verifier"]

        def fake_send(sess, req, **kwargs):
            self.assertIn(f"code_verifier={verifier}", req.body)
            return mock_send_value(get_bearer_token())

        with mock.patch("requests.sessions.Session.send", fake_send):
            request2 = self.factory.get(f"/authorize?state={state}")
            request2.session = request.session
            token = client.authorize_access_token(request2)
            self.assertEqual(token["access_token"], "a")

    def test_oauth2_authorize_code_verifier(self):
        request = self.factory.get("/login")
        request.session = self.factory.session

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            client_kwargs={"code_challenge_method": "S256"},
        )
        state = "foo"
        code_verifier = "bar"
        rv = client.authorize_redirect(
            request, "https://a.b/c", state=state, code_verifier=code_verifier
        )
        self.assertEqual(rv.status_code, 302)
        url = rv.get("Location")
        self.assertIn("state=", url)
        self.assertIn("code_challenge=", url)

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())

            request2 = self.factory.get(f"/authorize?state={state}")
            request2.session = request.session

            token = client.authorize_access_token(request2)
            self.assertEqual(token["access_token"], "a")

    def test_openid_authorize(self):
        request = self.factory.get("/login")
        request.session = self.factory.session
        secret_key = JsonWebKey.import_key("secret", {"kty": "oct", "kid": "f"})

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            jwks={"keys": [secret_key.as_dict()]},
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            client_kwargs={"scope": "openid profile"},
        )

        resp = client.authorize_redirect(request, "https://b.com/bar")
        self.assertEqual(resp.status_code, 302)
        url = resp.get("Location")
        self.assertIn("nonce=", url)
        query_data = dict(url_decode(urlparse.urlparse(url).query))

        token = get_bearer_token()
        token["id_token"] = generate_id_token(
            token,
            {"sub": "123"},
            secret_key,
            alg="HS256",
            iss="https://i.b",
            aud="dev",
            exp=3600,
            nonce=query_data["nonce"],
        )
        state = query_data["state"]
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(token)

            request2 = self.factory.get(f"/authorize?state={state}&code=foo")
            request2.session = request.session

            token = client.authorize_access_token(request2)
            self.assertEqual(token["access_token"], "a")
            self.assertIn("userinfo", token)
            self.assertEqual(token["userinfo"]["sub"], "123")

    def test_oauth2_access_token_with_post(self):
        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )
        payload = {"code": "a", "state": "b"}

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())
            request = self.factory.post("/token", data=payload)
            request.session = self.factory.session
            request.session["_state_dev_b"] = {"data": {}}
            token = client.authorize_access_token(request)
            self.assertEqual(token["access_token"], "a")

    def test_with_fetch_token_in_oauth(self):
        def fetch_token(name, request):
            return {"access_token": name, "token_type": "bearer"}

        oauth = OAuth(fetch_token=fetch_token)
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        def fake_send(sess, req, **kwargs):
            self.assertEqual(sess.token["access_token"], "dev")
            return mock_send_value(get_bearer_token())

        with mock.patch("requests.sessions.Session.send", fake_send):
            request = self.factory.get("/login")
            client.get("/user", request=request)

    def test_with_fetch_token_in_register(self):
        def fetch_token(request):
            return {"access_token": "dev", "token_type": "bearer"}

        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
            fetch_token=fetch_token,
        )

        def fake_send(sess, req, **kwargs):
            self.assertEqual(sess.token["access_token"], "dev")
            return mock_send_value(get_bearer_token())

        with mock.patch("requests.sessions.Session.send", fake_send):
            request = self.factory.get("/login")
            client.get("/user", request=request)

    def test_request_without_token(self):
        oauth = OAuth()
        client = oauth.register(
            "dev",
            client_id="dev",
            client_secret="dev",
            api_base_url="https://i.b/api",
            access_token_url="https://i.b/token",
            authorize_url="https://i.b/authorize",
        )

        def fake_send(sess, req, **kwargs):
            auth = req.headers.get("Authorization")
            self.assertIsNone(auth)
            resp = mock.MagicMock()
            resp.text = "hi"
            resp.status_code = 200
            return resp

        with mock.patch("requests.sessions.Session.send", fake_send):
            resp = client.get("/api/user", withhold_token=True)
            self.assertEqual(resp.text, "hi")
            self.assertRaises(OAuthError, client.get, "https://i.b/api/user")



import json
import time

from django.http import JsonResponse
from django.test import override_settings

from authlib.common.encoding import to_unicode
from authlib.common.urls import add_params_to_uri
from authlib.integrations.django_oauth1 import ResourceProtector
from authlib.oauth1.rfc5849 import signature
from tests.util import read_file_path

from .models import Client
from .models import TokenCredential
from .models import User
from .oauth1_server import TestCase


class ResourceTest(TestCase):
    def create_route(self):
        require_oauth = ResourceProtector(Client, TokenCredential)

        @require_oauth()
        def handle(request):
            user = request.oauth1_credential.user
            return JsonResponse(dict(username=user.username))

        return handle

    def prepare_data(self):
        user = User(username="foo")
        user.save()

        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            default_redirect_uri="https://a.b",
        )
        client.save()

        tok = TokenCredential(
            user_id=user.pk,
            client_id=client.client_id,
            oauth_token="valid-token",
            oauth_token_secret="valid-token-secret",
        )
        tok.save()

    def test_invalid_request_parameters(self):
        self.prepare_data()
        handle = self.create_route()
        url = "/user"

        # case 1
        request = self.factory.get(url)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_consumer_key", data["error_description"])

        # case 2
        request = self.factory.get(add_params_to_uri(url, {"oauth_consumer_key": "a"}))
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "invalid_client")

        # case 3
        request = self.factory.get(
            add_params_to_uri(url, {"oauth_consumer_key": "client"})
        )
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_token", data["error_description"])

        # case 4
        request = self.factory.get(
            add_params_to_uri(url, {"oauth_consumer_key": "client", "oauth_token": "a"})
        )
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "invalid_token")

        # case 5
        request = self.factory.get(
            add_params_to_uri(
                url, {"oauth_consumer_key": "client", "oauth_token": "valid-token"}
            )
        )
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_timestamp", data["error_description"])

    @override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["PLAINTEXT"]})
    def test_plaintext_signature(self):
        self.prepare_data()
        handle = self.create_route()
        url = "/user"

        # case 1: success
        auth_header = (
            'OAuth oauth_consumer_key="client",'
            'oauth_signature_method="PLAINTEXT",'
            'oauth_token="valid-token",'
            'oauth_signature="secret&valid-token-secret"'
        )
        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertIn("username", data)

        # case 2: invalid signature
        auth_header = auth_header.replace("valid-token-secret", "invalid")
        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "invalid_signature")

    def test_hmac_sha1_signature(self):
        self.prepare_data()
        handle = self.create_route()
        url = "/user"

        params = [
            ("oauth_consumer_key", "client"),
            ("oauth_token", "valid-token"),
            ("oauth_signature_method", "HMAC-SHA1"),
            ("oauth_timestamp", str(int(time.time()))),
            ("oauth_nonce", "hmac-sha1-nonce"),
        ]
        base_string = signature.construct_base_string(
            "GET", "http://testserver/user", params
        )
        sig = signature.hmac_sha1_signature(base_string, "secret", "valid-token-secret")
        params.append(("oauth_signature", sig))
        auth_param = ",".join([f'{k}="{v}"' for k, v in params])
        auth_header = "OAuth " + auth_param

        # case 1: success
        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertIn("username", data)

        # case 2: exists nonce
        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "invalid_nonce")

    @override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["RSA-SHA1"]})
    def test_rsa_sha1_signature(self):
        self.prepare_data()
        handle = self.create_route()

        url = "/user"

        params = [
            ("oauth_consumer_key", "client"),
            ("oauth_token", "valid-token"),
            ("oauth_signature_method", "RSA-SHA1"),
            ("oauth_timestamp", str(int(time.time()))),
            ("oauth_nonce", "rsa-sha1-nonce"),
        ]
        base_string = signature.construct_base_string(
            "GET", "http://testserver/user", params
        )
        sig = signature.rsa_sha1_signature(
            base_string, read_file_path("rsa_private.pem")
        )
        params.append(("oauth_signature", sig))
        auth_param = ",".join([f'{k}="{v}"' for k, v in params])
        auth_header = "OAuth " + auth_param

        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertIn("username", data)

        # case: invalid signature
        auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
        auth_header = "OAuth " + auth_param
        request = self.factory.get(url, HTTP_AUTHORIZATION=auth_header)
        resp = handle(request)
        data = json.loads(to_unicode(resp.content))
        self.assertEqual(data["error"], "invalid_signature")



from flask import json

from authlib.oauth2.rfc6749.grants import ClientCredentialsGrant
from authlib.oauth2.rfc7523 import JWTBearerClientAssertion
from authlib.oauth2.rfc7523 import client_secret_jwt_sign
from authlib.oauth2.rfc7523 import private_key_jwt_sign
from tests.util import read_file_path

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class JWTClientCredentialsGrant(ClientCredentialsGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        JWTBearerClientAssertion.CLIENT_AUTH_METHOD,
    ]


class JWTClientAuth(JWTBearerClientAssertion):
    def validate_jti(self, claims, jti):
        return True

    def resolve_client_public_key(self, client, headers):
        if headers["alg"] == "RS256":
            return read_file_path("jwk_public.json")
        return client.client_secret


class ClientCredentialsTest(TestCase):
    def prepare_data(self, auth_method, validate_jti=True):
        server = create_authorization_server(self.app)
        server.register_grant(JWTClientCredentialsGrant)
        server.register_client_auth_method(
            JWTClientAuth.CLIENT_AUTH_METHOD,
            JWTClientAuth("https://localhost/oauth/token", validate_jti),
        )

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="credential-client",
            client_secret="credential-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
                "grant_types": ["client_credentials"],
                "token_endpoint_auth_method": auth_method,
            }
        )
        db.session.add(client)
        db.session.commit()

    def test_invalid_client(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD)
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_invalid_jwt(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": client_secret_jwt_sign(
                    client_secret="invalid-secret",
                    client_id="credential-client",
                    token_endpoint="https://localhost/oauth/token",
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_not_found_client(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": client_secret_jwt_sign(
                    client_secret="credential-secret",
                    client_id="invalid-client",
                    token_endpoint="https://localhost/oauth/token",
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_not_supported_auth_method(self):
        self.prepare_data("invalid")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": client_secret_jwt_sign(
                    client_secret="credential-secret",
                    client_id="credential-client",
                    token_endpoint="https://localhost/oauth/token",
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_client_secret_jwt(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": client_secret_jwt_sign(
                    client_secret="credential-secret",
                    client_id="credential-client",
                    token_endpoint="https://localhost/oauth/token",
                    claims={"jti": "nonce"},
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_private_key_jwt(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": private_key_jwt_sign(
                    private_key=read_file_path("jwk_private.json"),
                    client_id="credential-client",
                    token_endpoint="https://localhost/oauth/token",
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_not_validate_jti(self):
        self.prepare_data(JWTBearerClientAssertion.CLIENT_AUTH_METHOD, False)

        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "client_credentials",
                "client_assertion_type": JWTBearerClientAssertion.CLIENT_ASSERTION_TYPE,
                "client_assertion": client_secret_jwt_sign(
                    client_secret="credential-secret",
                    client_id="credential-client",
                    token_endpoint="https://localhost/oauth/token",
                ),
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)



from authlib.oauth2.rfc6749.grants import ImplicitGrant

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class ImplicitTest(TestCase):
    def prepare_data(self, is_confidential=False, response_type="token"):
        server = create_authorization_server(self.app)
        server.register_grant(ImplicitGrant)
        self.server = server

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        if is_confidential:
            client_secret = "implicit-secret"
            token_endpoint_auth_method = "client_secret_basic"
        else:
            client_secret = ""
            token_endpoint_auth_method = "none"

        client = Client(
            user_id=user.id,
            client_id="implicit-client",
            client_secret=client_secret,
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["http://localhost/authorized"],
                "scope": "profile",
                "response_types": [response_type],
                "grant_types": ["implicit"],
                "token_endpoint_auth_method": token_endpoint_auth_method,
            }
        )
        self.authorize_url = (
            "/oauth/authorize?response_type=token&client_id=implicit-client"
        )
        db.session.add(client)
        db.session.commit()

    def test_get_authorize(self):
        self.prepare_data()
        rv = self.client.get(self.authorize_url)
        self.assertEqual(rv.data, b"ok")

    def test_confidential_client(self):
        self.prepare_data(True)
        rv = self.client.get(self.authorize_url)
        self.assertIn(b"invalid_client", rv.data)

    def test_unsupported_client(self):
        self.prepare_data(response_type="code")
        rv = self.client.get(self.authorize_url)
        self.assertIn(b"unauthorized_client", rv.data)

    def test_invalid_authorize(self):
        self.prepare_data()
        rv = self.client.post(self.authorize_url)
        self.assertIn("#error=access_denied", rv.location)

        self.server.scopes_supported = ["profile"]
        rv = self.client.post(self.authorize_url + "&scope=invalid")
        self.assertIn("#error=invalid_scope", rv.location)

    def test_authorize_token(self):
        self.prepare_data()
        rv = self.client.post(self.authorize_url, data={"user_id": "1"})
        self.assertIn("access_token=", rv.location)

        url = self.authorize_url + "&state=bar&scope=profile"
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("access_token=", rv.location)
        self.assertIn("state=bar", rv.location)
        self.assertIn("scope=profile", rv.location)

    def test_token_generator(self):
        m = "tests.flask.test_oauth2.oauth2_server:token_generator"
        self.app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
        self.prepare_data()
        rv = self.client.post(self.authorize_url, data={"user_id": "1"})
        self.assertIn("access_token=i-implicit.1.", rv.location)



import os

from authlib.integrations.django_oauth1 import CacheAuthorizationServer
from tests.django_helper import TestCase as _TestCase

from .models import Client
from .models import TokenCredential


class TestCase(_TestCase):
    def setUp(self):
        super().setUp()
        os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"

    def tearDown(self):
        os.environ.pop("AUTHLIB_INSECURE_TRANSPORT")
        super().tearDown()

    def create_server(self):
        return CacheAuthorizationServer(Client, TokenCredential)
